#include "alsa.h"

#include <stdio.h>
#include <stdlib.h>
#include <alsa/asoundlib.h>

typedef struct {
  alsa_out common;
  snd_pcm_t *h;
} alsa_out_internal;

#define A do { if (err < 0) { fprintf(stderr, "%s:%d: alsa error '%s'\n", __FILE__, __LINE__, snd_strerror(err)); abort(); } } while (0)

static snd_pcm_t *open_alsa(int samplerate, int channels)
{
  snd_pcm_t *h;
  snd_pcm_hw_params_t *hw_params;
  snd_pcm_sw_params_t *sw_params;
  snd_pcm_uframes_t   period_frames;
  snd_pcm_uframes_t   buffer_frames;
  unsigned int        rate;
  snd_pcm_format_t    format;
  int err;

  format = SND_PCM_FORMAT_S16_LE; //FLOAT_LE;
  err = snd_pcm_open(&h, "default", SND_PCM_STREAM_PLAYBACK, 0); A;
  snd_pcm_hw_params_alloca(&hw_params);
  snd_pcm_sw_params_alloca(&sw_params);
  err = snd_pcm_hw_params_any(h, hw_params); A;
  err = snd_pcm_hw_params_set_access(h, hw_params, SND_PCM_ACCESS_RW_INTERLEAVED); A;
  err = snd_pcm_hw_params_set_format(h, hw_params, format); A;
  err = snd_pcm_hw_params_set_channels(h, hw_params, channels); A;
  rate = samplerate;
  err = snd_pcm_hw_params_set_rate_near(h, hw_params, &rate, 0); A;
  if (rate != samplerate) {printf("bad rate\n"); abort();}
  period_frames = 1024;
  buffer_frames = 1024*20;
  err = snd_pcm_hw_params_set_period_size_near(h,hw_params,&period_frames,0);A;
  err = snd_pcm_hw_params_set_buffer_size_near(h, hw_params, &buffer_frames);A;
  err = snd_pcm_hw_params(h, hw_params); A;

  return h;
}

void xrun(snd_pcm_t *h)
{
  snd_pcm_prepare(h);
}

void suspend(snd_pcm_t *h)
{
  int res;
  int loop = 0;
  while ((res = snd_pcm_resume(h)) == -EAGAIN && loop < 100) {
    struct timeval t;
    t.tv_sec = 0;
    t.tv_usec = 10 * 1000;
    select(0, 0, 0, 0, &t);
    loop++;
  }
  if (res < 0) snd_pcm_prepare(h);
}

static void put_alsa(snd_pcm_t *h, unsigned char *bb, int frames)
{
#define X do { fprintf(stderr, "%s:%s:%d: alsa error\n", __FILE__, __FUNCTION__, __LINE__); xrun(h); } while (0)
#define Y do { fprintf(stderr, "%s:%s:%d: alsa error\n", __FILE__, __FUNCTION__, __LINE__); suspend(h); } while (0)
  ssize_t r;
  size_t count = frames;
  int    channels = 2;

  while (count) {
    r = snd_pcm_writei(h, bb, count);
    if (r == -EAGAIN || (r >= 0 && (size_t)r < count)) {
      /* nothing */
    }
    else if (r == -EPIPE) { /*usleep(1000);*/ X; }
    else if (r == -ESTRPIPE) Y;
    else if (r < 0) abort();
    if (r > 0) {
      count -= r;
      bb += r * 4 * channels;
    }
  }
#undef X
#undef Y
}

static void receive_audio(void *_alsa_out, void *buffer, int size)
{
  alsa_out_internal *alsa_out = _alsa_out;
  printf("audio data %d\n", size);
  put_alsa(alsa_out->h, buffer, size/4);
}

alsa_out *new_alsa_out(void)
{
  alsa_out_internal *ret = calloc(1, sizeof(alsa_out_internal));
  if (ret == NULL) exit(1);

  ret->common.receive_audio = receive_audio;

  ret->h = open_alsa(48000, 2);

  return (alsa_out *)ret;
}
