/* http://www.wohnort.org/DAB/france.html */

#include <stdio.h>

struct {
  char *name;
  double freq;
} band3[] = {
{(char*)"5A", 174.928},
{(char*)"5B", 176.640},
{(char*)"5C", 178.352},
{(char*)"5D", 180.064},
{(char*)"6A", 181.936},
{(char*)"6B", 183.648},
{(char*)"6C", 185.360},
{(char*)"6D", 187.072},
{(char*)"7A", 188.928},
{(char*)"7B", 190.640},
{(char*)"7C", 192.352},
{(char*)"7D", 194.064},
{(char*)"8A", 195.936},     /* busy out3 */  /* busy out1 */ /* busy out2 (too loud) */
{(char*)"8B", 197.648},     /* busy out2 */
{(char*)"8C", 199.360},
{(char*)"8D", 201.072},     /* busy out3 */ /* busy out2 (loud) */
{(char*)"9A", 202.928},
{(char*)"9B", 204.640},
{(char*)"9C", 206.352},
{(char*)"9D", 208.064},     /* busy? out2 */
{(char*)"10A", 209.936},
{(char*)"10B", 211.648},
{(char*)"10C", 213.360},
{(char*)"10D", 215.072},
{(char*)"10N", 210.096},
{(char*)"11A", 216.928},    /* busy out3 */ /* busy out1 */ /* busy out2 (loud) */
{(char*)"11B", 218.640},    /* busy out2 */
{(char*)"11C", 220.352},    /* busy out3 */ /* busy out1 */ /* busy out2 */
{(char*)"11D", 222.064},
{(char*)"11N", 217.088},    /* busy out3 */ /* busy out1 */ /* busy out2 */
{(char*)"12A", 223.936},
{(char*)"12B", 225.648},    /* busy out3 */ /* busy out1 */ /* busy out2 */
{(char*)"12C", 227.360},    /* busy out3 (low) */ /* busy out2 */
{(char*)"12D", 229.072},
{(char*)"12N", 224.096},    /* busy out2 */
{(char*)"13A", 230.784},
{(char*)"13B", 232.496},
{(char*)"13C", 234.208},
{(char*)"13D", 235.776},
{(char*)"13E", 237.488},
{(char*)"13F", 239.200}};

#include <uhd/usrp/multi_usrp.hpp>

void rx(double freq)
{
  uhd::usrp::multi_usrp::sptr usrp;
  uhd::rx_streamer::sptr rx_stream;
  uhd::tx_streamer::sptr tx_stream;
  std::string args = ""; //type=b200";
  uhd::device_addrs_t device_adds = uhd::device::find(args);

  if(device_adds.size() == 0) { printf("no b200?\n"); abort(); }

  args += ",num_recv_frames=256" ; 
  usrp = uhd::usrp::multi_usrp::make(args);
  usrp->set_rx_rate(2.048e6, 0);
  usrp->set_rx_freq(freq * 1000 * 1000, 0);

  ::uhd::gain_range_t gain_range = usrp->get_rx_gain_range(0);
  usrp->set_rx_gain(gain_range.stop(), 0);

  uhd::stream_args_t stream_args_rx("sc16", "sc16");
  stream_args_rx.channels.push_back(0);
  rx_stream = usrp->get_rx_stream(stream_args_rx);

  printf("rx stream max num samps: %d\n", (int)rx_stream->get_max_num_samps());
  usrp->set_rx_bandwidth(2.048e6, 0);
  usrp->set_time_now(uhd::time_spec_t(0.0));

      int i = 0;
      printf("RX Channel %u\n",i);
      std::cout << boost::format("Actual RX sample rate: %fMSps...") % (usrp->get_rx_rate(i)/1e6) << std::endl;
      std::cout << boost::format("Actual RX frequency: %fMHz...") % (usrp->get_rx_freq(i)/1e6) << std::endl;
      std::cout << boost::format("Actual RX gain: %f...") % (usrp->get_rx_gain(i)) << std::endl;
      std::cout << boost::format("Actual RX bandwidth: %fM...") % (usrp->get_rx_bandwidth(i)/1e6) << std::endl;
      std::cout << boost::format("Actual RX antenna: %s...") % (usrp->get_rx_antenna(i)) << std::endl;

  std::cout << boost::format("Device timestamp: %f...") % (usrp->get_time_now().get_real_secs()) << std::endl;

  uhd::stream_cmd_t cmd(uhd::stream_cmd_t::STREAM_MODE_START_CONTINUOUS);
  cmd.time_spec = usrp->get_time_now() + uhd::time_spec_t(0.05);
  cmd.stream_now = false; // start at constant delay
  rx_stream->issue_stream_cmd(cmd);

  FILE *f;
  char name[512];
  sprintf(name, "file%g.raw", freq);
  f = fopen(name, "w");
  for (i = 0; i < 4096; i++) {
  uhd::rx_metadata_t rx_md;
  unsigned char buf[1024*2*2];

  int n = rx_stream->recv(buf, 1024, rx_md);
  if (n != 1024) printf("ERROR: got %d samples instead of 1024\n", n);
  switch(rx_md.error_code){
  case uhd::rx_metadata_t::ERROR_CODE_NONE:
    break;
  case uhd::rx_metadata_t::ERROR_CODE_OVERFLOW:
    printf("[recv] USRP RX OVERFLOW!\n");
    break;
  case uhd::rx_metadata_t::ERROR_CODE_TIMEOUT:
    printf("[recv] USRP RX TIMEOUT!\n");
    break;
  default:
    printf("[recv] Unexpected error on RX, Error code: 0x%x\n",
           rx_md.error_code);
    break;
  }
  fwrite(buf, 1024*4, 1, f);
  }
  fclose(f);
}

int main(void)
{
  int i;

  for (i = 0; i < (int)(sizeof(band3)/sizeof(band3[0])); i++) {
    printf("%s: %g\n", band3[i].name, band3[i].freq);
    rx(band3[i].freq);
  }

  return 0;
}
